﻿using System.Linq.Expressions;
using Drive.Common.Domain;
using NHibernate;
using NHibernate.Linq;

namespace Drive.Common.Data
{
    public interface IAggregateRootLoader
    {
        Task<TAggregate> Load<TAggregate, TRecord>(Guid id)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord;

        Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(IEnumerable<Guid> ids)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord;

        Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(Expression<Func<TRecord, bool>> predicate)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord;

        Task<TAggregate> Load<TAggregate, TRecord>(int id)
            where TAggregate : ILegacyAggregateRoot<TRecord>
            where TRecord : ILegacyAggregateRootRecord;
    }

    public class AggregateRootLoader : IAggregateRootLoader
    {
        private readonly ISession _session;

        public AggregateRootLoader(
            ISession session)
        {
            _session = session;
        }

        public async Task<TAggregate> Load<TAggregate, TRecord>(Guid id)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord
        {
            TRecord record = await _session
                .GetAsync<TRecord>(id)
                .ConfigureAwait(false)
                ?? throw new ArgumentOutOfRangeException(nameof(id)); ;

            TAggregate result = (TAggregate?)Activator.CreateInstance(typeof(TAggregate), true)!;
            result.Restore(record);
            return result;
        }

        public async Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(IEnumerable<Guid> ids)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord
        {
            IEnumerable<TRecord> records = await _session
                .Query<TRecord>()
                .Where(x => ids.Contains(x.Id))
                .ToListAsync()
                .ConfigureAwait(false);

            List<TAggregate> aggregates = new();
            foreach (TRecord record in records)
            {
                TAggregate aggregate = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
                aggregate.Restore(record);
                aggregates.Add(aggregate);
            }

            return aggregates;
        }

        public async Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(Expression<Func<TRecord, bool>> predicate)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : IAggregateRootRecord
        {
            IEnumerable<TRecord> records = await _session
                .Query<TRecord>()
                .Where(predicate)
                .ToListAsync()
                .ConfigureAwait(false);

            List<TAggregate> aggregates = new();
            foreach (TRecord record in records)
            {
                TAggregate aggregate = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
                aggregate.Restore(record);
                aggregates.Add(aggregate);
            }

            return aggregates;
        }

        public async Task<TAggregate> Load<TAggregate, TRecord>(int id)
            where TAggregate : ILegacyAggregateRoot<TRecord>
            where TRecord : ILegacyAggregateRootRecord
        {
            TRecord record = await _session
                .GetAsync<TRecord>(id)
                .ConfigureAwait(false)
                ?? throw new ArgumentOutOfRangeException(nameof(id));

            TAggregate result = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
            result.Restore(record);
            return result;
        }
    }
}
