﻿using Drive.Common.Domain;
using NHibernate;

namespace Drive.Common.Data
{
    public interface IAggregateRootPersister
    {
        Task Persist<TRecord>(
            IAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : IAggregateRootRecord;
        Task Persist<TRecord>(
            ILegacyAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : ILegacyAggregateRootRecord;
        Task Persist<TRecord>(IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : IAggregateRootRecord;
        Task Delete<TRecord>(IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : IAggregateRootRecord;
        Task Delete<TRecord>(IAggregateRoot<TRecord> aggregate) where TRecord : IAggregateRootRecord;
    }

    public class AggregateRootPersister
        : IAggregateRootPersister
    {
        private readonly ISession _session;
        private readonly IClockProvider _clockProvider;

        public AggregateRootPersister(
            ISession session,
            IClockProvider clockProvider)
        {
            _session = session;
            _clockProvider = clockProvider;
        }

        public async Task Persist<TRecord>(
            IAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : IAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();
            IEnumerable<object> events = aggregate.GetEvents();

            if (context != null && record is IAuditedAggregateRoot root)
            {
                root.ModifiedById = context.DriveUserId;
            }

            if (record.ModifiedAt.HasValue && events.Any())
                record.ModifiedAt = _clockProvider.UtcNow;

            await _session
                .SaveAsync(record)
                .ConfigureAwait(false);
            await _session.FlushAsync()
                .ConfigureAwait(false);
        }

        public async Task Persist<TRecord>(
           IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : IAggregateRootRecord
        {
            foreach (IAggregateRoot<TRecord> aggregate in aggregates)
            {
                TRecord record = aggregate.GetRecord();
                IEnumerable<object> events = aggregate.GetEvents();

                if (record.ModifiedAt.HasValue && events.Any())
                    record.ModifiedAt = _clockProvider.UtcNow;

                await _session
                    .SaveAsync(record)
                    .ConfigureAwait(false);
            }

            await _session.FlushAsync()
                .ConfigureAwait(false);
        }

        public async Task Persist<TRecord>(
            ILegacyAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : ILegacyAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();
            var events = aggregate.GetEvents();

            if (record.ModifiedAt.HasValue && events.Any())
                record.ModifiedAt = _clockProvider.UtcNow;

            if (context != null && record is ILegacyAuditedAggregateRoot root)
            {
                root.ModifiedById = context.DriveUserId;
            }

            await _session
                .SaveAsync(record)
                .ConfigureAwait(false);
            await _session.FlushAsync()
                .ConfigureAwait(false);
        }
        public async Task Delete<TRecord>(
            IAggregateRoot<TRecord> aggregate) where TRecord : IAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();

            await _session
                .DeleteAsync(record)
                .ConfigureAwait(false);
            await _session.FlushAsync()
                .ConfigureAwait(false);
        }

        public async Task Delete<TRecord>(
            IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : IAggregateRootRecord
        {
            foreach (IAggregateRoot<TRecord> aggregate in aggregates)
            {
                TRecord record = aggregate.GetRecord();

                await _session
                    .DeleteAsync(record)
                    .ConfigureAwait(false);
            }
            await _session.FlushAsync()
                .ConfigureAwait(false);
        }
    }
}
