﻿using NodaTime;

namespace Drive.Common.Domain
{
    public interface IAggregateRootRecord
    {
        Guid Id { get; }
        int CreatedById { get; }
        Instant CreatedAt { get; }
        Instant? ModifiedAt { get; set; }
    }

    public interface IAuditedAggregateRoot
    {
        int ModifiedById { get; set; }
    }

    public interface IAggregateRoot<TRecord> where TRecord : IAggregateRootRecord
    {
        IEnumerable<object> GetEvents();
        TRecord GetRecord();
        void Restore(TRecord record);
    }

    public abstract class AggregateRoot<TRecord>
        : IAggregateRoot<TRecord> where TRecord : IAggregateRootRecord
    {
        private readonly Dictionary<Type, Action<object>> _handlers = new();
        private readonly Dictionary<Type, Action<object>> _audit = new();
        private readonly List<object> _events = new();

#pragma warning disable CS8618 // Restore() should be called at construction
        protected TRecord _record;
#pragma warning restore CS8618

        protected void Register<T>(
            Action<T> when,
            Action<T, TRecord>? audit = null)
        {
            _handlers.Add(typeof(T), e => when((T)e));
            if (audit != null)
                _audit.Add(typeof(T), (e) => audit((T)e, _record!));
        }

        protected void Raise(object e)
        {
            Type type = e.GetType();
            if (_audit.ContainsKey(type))
                _audit[type](e);
            _handlers[type](e);
            _events.Add(e);
        }

        public IEnumerable<object> GetEvents()
        {
            return _events;
        }

        public TRecord GetRecord()
        {
            return _record!;
        }

        void IAggregateRoot<TRecord>.Restore(TRecord record)
        {
            _record = record;
        }
    }
}
