﻿using IdentityModel.Client;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using Microsoft.IdentityModel.Tokens;

namespace Drive.Common.Security.Token
{
    public class DiscoveryDocumentSecurityKeyResolver
        : ISecurityKeyResolver
    {
        private static readonly object _padlock = new();

        private readonly ILogger _logger;
        private readonly IOptionsMonitor<JwtBearerOptions> _jwtOptions;
        private readonly DriveAuthenticationOptions _authOptions;
        private IEnumerable<SecurityKey> _securityKeys;

        public DiscoveryDocumentSecurityKeyResolver(
            IOptionsMonitor<JwtBearerOptions> jwtOptions,
            IOptions<DriveAuthenticationOptions> authOptions,
            ILoggerFactory loggerFactory)
        {
            _jwtOptions = jwtOptions;
            _authOptions = authOptions.Value;
            _securityKeys = Enumerable.Empty<SecurityKey>();
            _logger = loggerFactory.CreateLogger("authentication-logger");
        }

        public IEnumerable<SecurityKey> GetSecurityKeys(string keyId)
        {
            List<SecurityKey> keys = FilterKeys(keyId);

            if (!keys.Any())
            {
                lock (_padlock)
                {
                    keys = FilterKeys(keyId); // try again as previous thread exited lock and may have rereshed the keys
                    if (!keys.Any())
                    {
                        _securityKeys = Task.Run(() => GetKeys()).GetAwaiter().GetResult();
                    }

                    keys = FilterKeys(keyId);
                    if (!keys.Any())
                        _logger.LogWarning($"Could not find key with key id {keyId}");
                }
            }

            return keys;
        }

        private async Task<IEnumerable<SecurityKey>> GetKeys()
        {
            JwtBearerOptions options = _jwtOptions.Get(JwtBearerDefaults.AuthenticationScheme);

            using HttpClient client = new();

            DiscoveryDocumentResponse disco = await client.GetDiscoveryDocumentAsync(
                new DiscoveryDocumentRequest
                {
                    Address = options.Authority,
                    Policy = new DiscoveryPolicy
                    {
                        AdditionalEndpointBaseAddresses = _authOptions.DiscoveryDocumentBaseAddresses.ToList(),
                    }
                });

            JsonWebKeySet keyset = JsonWebKeySet.Create(disco.KeySet!.RawData);

            return keyset
                .GetSigningKeys()
                .ToList();
        }

        private List<SecurityKey> FilterKeys(string keyId)
        {
            return _securityKeys
                .Where(x => x.KeyId == keyId)
                .ToList();
        }
    }
}
