﻿using System.Data;
using System.Data.Common;
using System.Reflection;
using Drive.Common.Data;
using Microsoft.Data.SqlClient;
using NHibernate;
using NHibernate.Cfg;
using NHibernate.Driver;
using Xunit;

namespace Drive.Common.Testing.Database
{
    public abstract class DatabaseTest : IDisposable
    {
        private const string CREATE_SCHEMA_SQL = @"
            IF NOT EXISTS (SELECT name FROM sys.schemas WHERE name = N'{0}')
                EXEC('CREATE SCHEMA {0}');";

        private readonly string _serverConnectionString = @"Data Source=(LocalDB)\MSSQLLocalDB";

        private readonly string _databaseName;
        private readonly string _connectionString;

        private readonly Configuration _configuration;
        private readonly ISessionFactory _sessionFactory;

        public DatabaseTest(string? serverConnectionString = null)
        {
            _serverConnectionString = serverConnectionString ?? _serverConnectionString;

            Type collectionDefinition = typeof(ICollectionFixture<>).MakeGenericType(GetType());

            Type? collection = AppDomain.CurrentDomain
                                        .GetAssemblies()
                                        .Where(x => !string.IsNullOrWhiteSpace(x.FullName) && x.FullName.Contains("Drive"))
                                        .SelectMany(x => x.ExportedTypes)
                                        .Where(x => collectionDefinition.IsAssignableFrom(x))
                                        .FirstOrDefault();

            string name = collection?.Assembly.GetName().Name ?? Assembly.GetExecutingAssembly().GetName().Name!;

            _databaseName = $"{name}_{GetType().Name}".Replace(".", "-");

            DropDatabase();
            CreateDatabase();
            _connectionString = $"{_serverConnectionString};Initial Catalog={_databaseName};Pooling=false";

            List<string> schemaNames = new();
            AddSchemaNamesToCreate(schemaNames);
            CreateSchemaObjects(schemaNames);

            CreateAuxiliaryObjects();

            _configuration = BuildConfiguration();
            AddMappings(_configuration);
            _sessionFactory = _configuration.BuildSessionFactory();

            AddInitialTestData();
        }

        protected virtual void AddSchemaNamesToCreate(List<string> schemaNames) { }
        protected virtual void CreateAuxiliaryObjects() { }
        protected virtual void AddMappings(Configuration configuration) { }
        protected virtual void AddInitialTestData() { }

        public SqlConnection OpenConnection(string? connectionString = null)
        {
            SqlConnection connection = new(connectionString ?? _connectionString);
            connection.Open();

            return connection;
        }

        public ISession OpenSession(
            DbConnection? connection = null)
        {
            ISessionBuilder builder = _sessionFactory
                .WithOptions();
            if (connection != null)
                builder = builder.Connection(connection);

            return builder.OpenSession();
        }

        public IStatelessSession OpenStatelessSession()
        {
            return _sessionFactory.OpenStatelessSession();
        }

        public void ResetData()
        {
            using (SqlConnection connection = OpenConnection())
            using (SqlTransaction transaction = connection.BeginTransaction())
            using (SqlCommand command = connection.CreateCommand())
            {
                command.Transaction = transaction;
                command.CommandType = CommandType.StoredProcedure;
                command.CommandText = "sp_MSForEachTable";

                // disable referential integrity
                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "ALTER TABLE ? NOCHECK CONSTRAINT ALL"));
                command.ExecuteNonQuery();

                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "SET QUOTED_IDENTIFIER ON; DELETE FROM ?"));
                command.ExecuteNonQuery();

                // enable referential integrity again
                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "ALTER TABLE ? WITH CHECK CHECK CONSTRAINT ALL"));
                command.ExecuteNonQuery();

                transaction.Commit();
            }

            AddInitialTestData();
        }

        private void CreateDatabase()
        {
            using SqlConnection connection = OpenConnection(_serverConnectionString);
            using SqlCommand command = connection.CreateCommand();

            command.CommandText = $@"
CREATE DATABASE [{_databaseName}]
ON PRIMARY      (NAME=[{_databaseName}_data], FILENAME = 'C:\temp\{_databaseName}_data.mdf')
LOG ON          (NAME=[{_databaseName}_log], FILENAME = 'C:\temp\{_databaseName}_log.ldf')";
            command.ExecuteNonQuery();
        }

        private void DropDatabase()
        {
            using SqlConnection connection = OpenConnection(_serverConnectionString);
            using SqlCommand command = connection.CreateCommand();

            command.CommandText = $@"
IF EXISTS (SELECT name FROM sys.databases WHERE name = N'{_databaseName}')
    BEGIN
        ALTER DATABASE [{_databaseName}] SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
        DROP DATABASE [{_databaseName}];
    END";
            command.ExecuteNonQuery();
        }

        private void CreateSchemaObjects(List<string> schemaNames)
        {
            using SqlConnection connection = OpenConnection();
            using SqlCommand command = connection.CreateCommand();

            foreach (string schema in schemaNames)
            {
                command.CommandText = string.Format(CREATE_SCHEMA_SQL, schema);
                command.ExecuteNonQuery();
            }
        }

        private Configuration BuildConfiguration()
        {
            Configuration configuration = new();
            configuration.DataBaseIntegration(c =>
            {
                c.ConnectionString = _connectionString;
                c.Driver<MicrosoftDataSqlClientDriver>();
                c.Dialect<DriveMsSql2012Dialect>();
                c.SchemaAction = SchemaAutoAction.Update;
            });
            configuration.SetProperty(NHibernate.Cfg.Environment.UseConnectionOnSystemTransactionPrepare, "false");

            return configuration;
        }

        public void Dispose()
        {
            _sessionFactory.Dispose();
            DropDatabase();
        }
    }
}