﻿using System.Security.Claims;
using Drive.Common.Query;
using Drive.Common.Security;
using Microsoft.AspNetCore.Authentication;
using Microsoft.Extensions.Logging;
using NodaTime;

namespace Drive.Common.Web
{
    public class ClaimsTransformer :
        IClaimsTransformation
    {
        private readonly ICacheClaimsPrincipals _cache;
        private readonly Lazy<IQueryHandler<string, IDriveUser?>> _userProvider;
        private readonly IClockProvider _clockProvider;
        private readonly ILogger _logger;

        public ClaimsTransformer(
            ICacheClaimsPrincipals cache,
            Lazy<IQueryHandler<string, IDriveUser?>> userProvider,
            IClockProvider clockProvider,
            ILoggerFactory loggerFactory)
        {
            _cache = cache;
            _userProvider = userProvider;
            _clockProvider = clockProvider;
            _logger = loggerFactory.CreateLogger("authentication-logger");
        }

        public Task<ClaimsPrincipal> TransformAsync(ClaimsPrincipal incomingPrincipal)
        {
            _logger.LogInformation("claims transform started");

            if (incomingPrincipal.Identity is null || !incomingPrincipal.Identity.IsAuthenticated)
                return Task.FromResult(incomingPrincipal);

            _logger.LogInformation("authenticated principal found");

            Instant tokenExpiration = GetTokenValidTo(incomingPrincipal);
            string username = ExtractUsername(incomingPrincipal);

            ClaimsPrincipal? cachedPrincipal = _cache.Get(username);
            if (cachedPrincipal != null)
                return Task.FromResult(cachedPrincipal);

            IDriveUser? user = _userProvider.Value.Retrieve(username);
            if (user == null)
                return Task.FromResult(incomingPrincipal);

            _logger.LogInformation("user found, get claims and map roles to claims");

            List<Claim> claims = [];
            AddUserDetailClaims(claims, user);
            AddRolesToClaimSet(claims, user);

            ClaimsPrincipal transformed = CreateTransformedPrincipal(claims, incomingPrincipal);
            _cache.Set(username, transformed, tokenExpiration);

            return Task.FromResult(transformed);
        }

        private Instant GetTokenValidTo(ClaimsPrincipal incomingPrincipal)
        {
            Instant validTo = _clockProvider.UtcNow;

            Claim? expiration = incomingPrincipal.FindFirst("exp");
            if (expiration != null)
                validTo = ConvertExpirationClaimToValidToDate(expiration);

            return validTo;
        }

        private static Instant ConvertExpirationClaimToValidToDate(Claim expiration)
        {
            return Instant.FromUnixTimeSeconds(long.Parse(expiration.Value));
        }

        private static string ExtractUsername(ClaimsPrincipal incomingPrincipal)
        {
            string? email = incomingPrincipal.FindFirst(ClaimTypes.Email)?.Value;
            return email != null ? email[..email.IndexOf('@')] : string.Empty;
        }

        private static void AddUserDetailClaims(List<Claim> claims, IDriveUser user)
        {
            claims.AddRange(new[]
            {
                new Claim(ClaimTypes.Name, user.Code),
                new Claim(DriveClaimTypes.UserId, user.Id.ToString()),
                new Claim(DriveClaimTypes.UserDescription, user.Description.ToString()),
                new Claim(DriveClaimTypes.UserBranchId, user.StaffBranchId.ToString()),
                new Claim(DriveClaimTypes.StaffId, user.StaffId.ToString())
            });
        }

        private static void AddRolesToClaimSet(List<Claim> claims, IDriveUser user)
        {
            IEnumerable<Claim> rolesForUser = GetRoleClaims(user);
            claims.AddRange(rolesForUser);
        }

        private static IEnumerable<Claim> GetRoleClaims(IDriveUser user)
        {
            return user.SecurityRoles.Select(role => new Claim(ClaimTypes.Role, role));
        }

        private static ClaimsPrincipal CreateTransformedPrincipal(List<Claim> claims, ClaimsPrincipal incomingPrincipal)
        {
            ClaimsIdentity identity = new(claims, incomingPrincipal.Identity!.AuthenticationType);
            return new ClaimsPrincipal(identity);
        }
    }
}
