﻿using System.Reflection;
using Drive.Common.Command;
using Drive.Common.Web.Extensions;
using Lamar;
using NServiceBus.MessageRouting.RoutingSlips;

namespace Drive.Common.Web.Command
{
    public interface ICommandDispatcher
    {
        Task<IDispatchResult> Dispatch<T>(T command, SendOptions options, IEnumerable<string> destinations);

        Task<IDispatchResult> Dispatch(IEnumerable<object> command, SendOptions options);
    }

    public class CommandDispatcher
        : ICommandDispatcher
    {
        private static readonly MethodInfo _validate = typeof(CommandDispatcher).GetMethod(nameof(CommandDispatcher.ValidateCommand), BindingFlags.Instance | BindingFlags.NonPublic)!;

        private readonly IContainer _container;
        private readonly IMessageSession _messageSession;

        public CommandDispatcher(
            IContainer container,
            IMessageSession messageSession)
        {
            _container = container;
            _messageSession = messageSession;
        }

        public async Task<IDispatchResult> Dispatch<T>(T command, SendOptions options, IEnumerable<string> destinations)
        {
            IDictionary<string, IEnumerable<string>> validationErrors = new Dictionary<string, IEnumerable<string>>();
            ValidateCommand<T>(command, validationErrors);

            DispatchResult result = new(validationErrors);
            if (result.Success)
            {
                await _messageSession
                .Route(
                    command!,
                    options,
                    destinations.ToArray())
                .ConfigureAwait(false);
            }

            return result;
        }

        public async Task<IDispatchResult> Dispatch(IEnumerable<object> commands, SendOptions options)
        {
            IDictionary<string, IEnumerable<string>> validationErrors = new Dictionary<string, IEnumerable<string>>();

            foreach (object command in commands)
            {
                _validate
                    .MakeGenericMethod(command.GetType())
                    .Invoke(this, new[] { command, validationErrors });
            }

            DispatchResult result = new(validationErrors);
            if (result.Success)
            {
                foreach (var command in commands)
                    await _messageSession
                        .Send(command, options)
                        .ConfigureAwait(false);
            }

            return result;
        }

        private void ValidateCommand<T>(
            T command,
            IDictionary<string, IEnumerable<string>> validationErrors)
        {
            Type typedValidator = typeof(ICommandValidator<>).MakeGenericType(command!.GetType());
            var validators = _container.GetAllInstances(typedValidator);
            foreach (var validator in validators)
            {
                IDictionary<string, IEnumerable<string>> validationResult = ((ICommandValidator<T>)validator).Validate(command);
                MergeValidationResult(validationErrors, validationResult);
            }
        }

        private static void MergeValidationResult(
            IDictionary<string, IEnumerable<string>> allErrors,
            IDictionary<string, IEnumerable<string>> newErrors)
        {
            foreach (string key in newErrors.Keys)
            {
                if (allErrors.TryGetValue(key, out IEnumerable<string>? value))
                    allErrors[key] = value.Concat(newErrors[key]);
                else
                    allErrors.Add(key, newErrors[key]);
            }
        }
    }
}
