﻿using Drive.Common.Extensions;
using FluentAssertions;
using Xunit;

namespace Drive.Common.UnitTest
{
    public class StringExtensionsSuite
    {
        [Fact]
        public void empty_string_returns_no_tokens()
        {
            //
            var empty = string.Empty;

            //
            IEnumerable<string> tokens = empty.Tokenize();

            //
            tokens.Should().HaveCount(0, "the source string was empty");
        }

        [Fact]
        public void string_with_no_spaces_returns_single_token()
        {
            //
            var aSingleToken = "asingletoken";

            //
            IEnumerable<string> tokens = aSingleToken.Tokenize();

            //
            tokens.Should().HaveCount(1, "the source string contains no spaces");
            tokens.First().Should().Be(aSingleToken, "the source string contains no spaces so is itself the only token");
        }

        [Fact]
        public void string_with_spaces_returns_correct_tokens()
        {
            //
            var source = "a source with five tokens";

            //
            IEnumerable<string> tokens = source.Tokenize();

            //
            tokens.Should().HaveCount(5, "the source string contains 5 tokens");
            tokens.Should().BeEquivalentTo(new string[] { "a", "source", "with", "five", "tokens"} , "those are the tokens contained in the source");
        }

        [Fact]
        public void string_starting_with_whitespace_returns_correct_tokens()
        {
            //
            var source = "         two tokens";

            //
            IEnumerable<string> tokens = source.Tokenize();

            //
            tokens.Should().HaveCount(2, "the source string contains 2 tokens");
            tokens.Should().BeEquivalentTo(new string[] { "two", "tokens" }, "those are the tokens contained in the source");
        }

        [Fact]
        public void string_ending_with_whitespace_returns_correct_tokens()
        {
            //
            var source = "two tokens         ";

            //
            IEnumerable<string> tokens = source.Tokenize();

            //
            tokens.Should().HaveCount(2, "the source string contains 2 tokens");
            tokens.Should().BeEquivalentTo(new string[] { "two", "tokens" }, "those are the tokens contained in the source");
        }

        [Fact]
        public void tokens_separated_by_large_amounts_of_whitespace_returns_correct_tokens()
        {
            //
            var source = "two         tokens";

            //
            IEnumerable<string> tokens = source.Tokenize();

            //
            tokens.Should().HaveCount(2, "the source string contains 2 tokens");
            tokens.Should().BeEquivalentTo(new string[] { "two", "tokens" }, "those are the tokens contained in the source");
        }

        [Fact]
        public void full_text_cleanse_converts_underscore_to_space()
        {
            //
            var source = "two_tokens";

            //
            string result = source.FullTextCleanse();

            //
            result.Should().Be("two tokens");
        }

        [Fact]
        public void full_text_cleanse_converts_hyphen_to_space()
        {
            //
            var source = "two-tokens";

            //
            string result = source.FullTextCleanse();

            //
            result.Should().Be("two tokens");
        }

        [Fact]
        public void full_text_cleanse_replaces_special_characters()
        {
            //
            var source = "two$%^tokens";

            //
            string result = source.FullTextCleanse();

            //
            result.Should().Be("twotokens");
        }

        [Fact]
        public void full_text_cleanse_maintains_spaces_around_special_characters()
        {
            //
            var source = "two & tokens";

            //
            string result = source.FullTextCleanse();

            //
            result.Should().Be("two  tokens");
        }

        [Fact]
        public void full_text_cleanse_handles_separators_and_special_characters()
        {
            //
            var source = "two_&-&*@# tokens/_";

            //
            string result = source.FullTextCleanse();

            //
            result.Should().Be("two   tokens ");
        }
    }
}
